/* Copyright 2001 Manuel Arriaga
 * 
 * 
 * This program is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <errno.h>
#include <setjmp.h>

#define __USE_GNU 1 /* for access to get_current_dir_name() */

#include <unistd.h>
#include <dirent.h>
#include <pwd.h>
#include <sys/stat.h>
#include <dlfcn.h>

#define UNLINK  1
#define NOTHING 0
#define YES 1
#define NO 0


/* You probably don't need to change this, unless you spend _lots_ of time deleting
 * files with the same name in the same dir, and  you have that dir covered by libtrash.
 * (Even then, it is _NOT_ necessary.)
 */

#define DEF_DIGITS 3


/* ---------------------------------- */

/* Configuration 
 * 
 * (Except for DEBUG, these are just compile-time defaults; all of them can
 * be changed at run-time by setting the appropriate environmental
 * variables): */


/* This macro defines whether we should output error messages to stderr or run silently. You should either
 * leave it commented out (to disable the messages) or uncomment it (to have messages printed to stderr):
 * (Note: "uncomment" implies removing the leading slash and star and trailing star and slash.)
 */

/*#define DEBUG*/


/* This macro sets the default name of the directory in the user's home directory in which "deleted"
 * files will be stored. It mustn't contain any slashes. This setting can be overriden
 by the environmental variable TRASH_CAN. */

#define TRASH_CAN "Trash"

/* This macro defines what should be done in the (unlikely!) case of failure: should we just
 *  call glibc's unlink() using the argument we were passed or do nothing at
 * all? (Note: There is one possible error situation in which we wouldn't
 * even be able to invoke glibc's unlink(): if the call to dlsym() fails. The
 * other possible errors are caused either by allocation failures, the
 * impossibility to determine information about the user or by the fact that
 * we were unable to create a file/directory. In any of these cases, we can
 * always just pass our argument to the "real" unlink(), if that is what the
 * user prefers.)
 * 
 * Please set it to either UNLINK or NOTHING (the latter option leads to a return value of -1
 * if an error occurs, while in the other case we just return the return value of glibc's unlink()).
 * This setting can be overriden by the  environmental variable TRASH_FAIL. 
 */

#define TRASH_FAIL NOTHING

/* This macro defines what we will do if asked to unlink() a file under the user's TRASH_CAN.
 * There is one thing we must refuse to do: handle it normally (i.e.,
 * storing a copy of it under TRASH_CAN/TRASH_CAN/). The file would never
 * be deleted and we would get a chain of TRASH_CAN directories under your
 * TRASH_CAN. So we can either call the "real" unlink() - by setting this
 * macro to UNLINK -, or do NOTHING. Please note that, if you set this
 * macro to UNLINK, you will _NOT_ be fully protected from devastating
 * commands such as:
 * 
 * joe@localhost:~$ rm -rf *
 * 
 * In this case, libtrash would store a copy of all "deleted"
 * files under your TRASH_CAN, but would still allow the removal of all of
 * them from your TRASH_CAN. (Depending on the order in which rm deletes
 * files under your home dir, the consequences might vary: if it only tried
 * to unlink() the files under TRASH_CAN _after_ having "deleted" all the
 * others, you would lose everything; if, on the other hand, it first
 * unlink()ed the files under TRASH_CAN and then proceeded to the others,
 * you would just lose the files you had stored under TRASH_CAN itself). 
 *
 * In other words, you probably want to set it to NOTHING, and use the provided
 * script (realrm.sh) to forcefully remove files under your TRASH_CAN. This setting
 * can be overriden by the environmental variable TRASH_RECURSE.
 */

#define TRASH_RECURSE NOTHING


/* This macro defines whether hidden files (or files under a hidden directory) should be handled
 * normally or really unlink()ed. Usually, hidden files are either temporary or configuration files.
 * Set to either YES or NO. YES tells libtrash to "ignore" such files, i.e., to really unlink() them.
 * Choosing NO causes them to be handled "normally", i.e., rename()d instead of really unlink()ed.
 * This setting can be overriden by the environmental variable TRASH_IGNORE_HIDDEN.
 */

#define TRASH_IGNORE_HIDDEN YES


/* This macro determines whether requests by users to unlink() files outside of their home
 * directory are intercepted, i.e., if such files are rename()d (i.e., moved to their TRASH_CAN)
 * instead of (really) unlink()ed. Of course, any of these things will only happen if the user 
 * has write-permission to those files.
 * 
 * If you set this variable to YES, you should take a look at TRASH_SYSTEM_ROOT below.
 * 
 * An example:
 * 
 * Joane tries to unlink() /common/latest-news.txt, to which she has write-permission.
 * If you set TRASH_GLOBAL to YES (and Joane hasn't overriden this setting with the environmental
 * variable TRASH_GLOBAL), /common/latest-news.txt will be moved to
 * /home/joane/Trash/SYSTEM_ROOT/common/latest-news.txt, assuming that TRASH_SYSTEM_ROOT is "SYSTEM_ROOT".
 * If TRASH_CAN had the value NO, /common/latest-news.txt would be unlink()ed using glibc's unlink(), and
 * no copy would be stored in Joane's TRASH_CAN. If Joane lacked write-permission to this file in the first
 * place, her request would naturally fail.
 * 
 * NOTES:
 * 
 * 
 * 1- This _DOESN'T_ apply to requests from root to unlink() files under one of the "protected"
 * directories listed in unremovable_dirs: if the super user lists "/lib" in unremovable_dirs,
 * and sets TRASH_GLOBAL, trying to unlink(), e.g., /lib/libc.so.6 will still fail: the original
 * file (actually, it is a symlink) will be left in the same place, and it won't be copied to her
 * TRASH_CAN. If TRASH_GLOBAL hadn't been set, exactly the same thing would have happened: this
 * request would have been ignored.
 * 
 * 2- /tmp is always ignored: requests to unlink() files under /tmp are passed on to glibc's
 * unlink() independently of the value of TRASH_GLOBAL. It doesn't make any sense to allow
 * storing back-up copies of temporary files; besides, this would make your TRASH_CAN unbrowsable
 * due to the quantity of litter (pun intended). If, for some unbelievable reason, root decides to
 * make /tmp unremovable (by listing it in unremovable_dirs), this still holds: requests to unlink()
 * files under /tmp are _ALWAYS_ respected, i.e., passed to the "real" unlink().
 */

#define TRASH_GLOBAL NO


/* This macro sets the default name of the directory under TRASH_CAN which will serve as "root"
 * if you have set TRASH_GLOBAL above (if TRASH_GLOBAL is set to NO, this value is ignored).
 * E.g.: if you set this macro to "SYSTEM_ROOT", and TRASH_GLOBAL is set to YES, trying to unlink() 
 * /common/interesting-stuff/important.html will result in a copy of the deleted file in
 * /(YOUR_HOME_DIR)/(YOUR_TRASH_CAN)/SYSTEM_ROOT/common/interesting-stuff/important.html.
 * Just like TRASH_CAN, it musn't contain any slashes - it should be a plain directory name:
 */

#define TRASH_SYSTEM_ROOT "SYSTEM_ROOT"


/* This macro sets a list of directories under which files won't be unlink()ed even if the request
 * came from the super user. They won't be transferred to root's TRASH_CAN: these requests are
 * simply ignored and -1 is returned after each call to unlink(). This list can be overriden by the
 * environmental variable TRASH_UNREM_DIRS. (Notes: 1- This setting doesn't have any effect on 
 * actions performed by non-privileged users. 2- If root's home dir is listed here, files under it
 * will be treated as files under the other unremovable_dirs: requests to unlink() them will always
 * fail, and they will remain in their place (i.e., they will _NOT_ be moved to root's TRASH_CAN).
 * 3- If UNREM_DIRS and TMP_DIRS (see below) have a directory in common, we handle it as if
 * it were listed only in UNREM_DIRS.)
 * 
 * Set to the empty string ("") if you don't need this feature. Otherwise, fill in a semi-colon separated 
 * list of directories:
 */

#define UNREM_DIRS "/bin;/boot;/dev;/lib;/proc;/sbin;/usr"


/* This macro sets a list of directories under which temporary files are created. We don't
 * redirect requests to unlink() files under any of these directories, even if root has 
 * listed it in UNREM_DIRS.
 * 
 * This must be a semi-colon separated list of directories. It can be overriden by the
 * environmental variable TRASH_TMP_DIRS:
 * 
 * Fill in the empty string ("") if you don't need/want this exception:
 * 
 */

#define TMP_DIRS "/tmp;/var"



/* End of configuration - no need to change anything below this section. */


/* ---------------------------------- */

/* Helper functions (defined at the bottom):  */

static char *get_string_env(const char *name);

static int handle_trash_error(const char *pathname);

static int found_under_dir(const char *absolute_path, const char *dir_list);

static int graft_file(const char *tree, const char *branch, const char *what_to_cut);

static int dir_ok(const char *pathname, int *name_collision);

static void reformulate_new_path(char **new_path, char **first_null);

static int hidden_file(const char *absolute_path);

static void cleanse_path(char **absolute_path);

/* ---------------------------------- */


/* Global variables (many are set to their compile-time default values specified
 above, which are hold in automatic storage): */


/* Configuration-related: ----------------------------------------- */


/* Controls what should be the behaviour of the library in case an error occurs: */

static int trash_fail = TRASH_FAIL;

/* Controls whether hidden files (or files under hidden directories) should be ignored
 * (i.e., really unlink()ed) or handled normally:
 */

static int trash_ignore_hidden = TRASH_IGNORE_HIDDEN;

/* Controls what we do when asked to unlink() files under the user's TRASH_CAN: */

static int trash_recurse = TRASH_RECURSE;

/* Controls whether attempts to unlink() files outside of the user's home dir are intercepted
 * (see detailed explanation above): */

static int trash_global = TRASH_GLOBAL;

/* This points to a list of directories (separated by semi-colons) which contain
 files that we should refuse to unlink even if running as root (or NULL, if this
 feature should be disabled by default): */

static char default_unremovable_dirs[] = UNREM_DIRS;

static char *unremovable_dirs = default_unremovable_dirs;

/* This points to a semi-colon delimited list of directories under which glibc's unlink() 
 * is always used, because they hold temporary files: */

static char default_tmp_dirs[] = TMP_DIRS;

static char *tmp_dirs = default_tmp_dirs;


/* User-related: ------------------------------------------------ */


/* This one points to the user's name:  */

static char *username = NULL;

/* This one points to the user's home directory: */

static char *home = NULL;

/* Points to the absolute path of the directory in the user's home dir to which "deleted" files are moved: */

static char *absolute_trash_can = NULL;


/* If trash_global is set to YES, files outside of the user's home dir which are unlink()ed have
 * a copy of themselves stored under absolute_trash_system_root:
 */

static char *absolute_trash_system_root = NULL;


/* Others: ------------------------------------------------------ */


/* Used by _init() to signal to unlink() if an error occurred while initializing. If it is
 set to 1, unlink() does whatever is specified in trash_fail: */

static int failure = 0;

/* This is a pointer to the unlink() function found in glibc: */

static int (*unlink_ptr) (const char*) = NULL;

/* This is used to allow found_under_dir() to make unlink() return an error code without 
 * complicating the code (a questionable decision, I suppose):
 */

static jmp_buf env;

/* End of global variables. */

/* ---------------------------------- */




/* _init() is called automatically upon loading. It is used to gather necessary user information and
 preferences (if any), perform some tests and "turn off" libtrash if it finds any problems. This is done
 * by setting the variable failure, whose value unlink() checks before running. */

void _init()
{
   
   struct passwd *userinfo = NULL;
   
   char default_relative_trash_can[] = TRASH_CAN;
   
   char *relative_trash_can = default_relative_trash_can;
   
   
   char default_relative_trash_system_root[] = TRASH_SYSTEM_ROOT;
   
   char *relative_trash_system_root = default_relative_trash_system_root;
   
   
   char *tmp = NULL;
   
   
   /* See if we need to override any of the compile-time defaults with values set in the environment: 
    * get_string_env returns either a dynamically-allocated string or NULL, depending on whether it had
    * success or not. In case of success, we must free() that memory. The reason why we use it instead of
    * calling getenv() directly is explained on the top of that function.
    */
   
   
   /* 1- Name of trash can directory in user's home dir (_without_ slashes): */
   
   /* This memory will be free()d below, either after we compose absolute_trash_can or right away,
    if it contains a slash: */
   
   tmp = get_string_env("TRASH_CAN");
   
   if (tmp)
     {
	/* If the environ var TRASH_CAN contains any slashes, we ignore it. It is meant to be just a dir name. */
	
	if (strchr(tmp, '/'))
	  free(tmp);
	else 
	  relative_trash_can = tmp;
     }
   
   
   /* 2- trash_fail determines our behaviour in case of error: */
   
   tmp = getenv("TRASH_FAIL");
   
   if (tmp)
     {
	if (!strcmp(tmp, "UNLINK"))
	  trash_fail = UNLINK;
	else if (!strcmp(tmp, "NOTHING"))
	  trash_fail = NOTHING;
	
	/* If the environmental variable TRASH_FAIL has any other value, we use the compile-time
	 default.*/
     }
   
   
   /* 3- trash_ignore_hidden determines whether hidden files (or files under hidden directories) are
    * handled normally or ignored:
    */
   
   tmp = getenv("TRASH_IGNORE_HIDDEN");
   
   if (tmp)
     {
	if (!strcmp(tmp, "YES"))
	  trash_ignore_hidden = YES;
	else if (!strcmp(tmp, "NO"))
	  trash_ignore_hidden = NO;
	/* If the environmental variable TRASH_IGNORE_HIDDEN has any other value, we use the compile_time
	 default. */
     }
   
   
   /* 4- trash_recurse determines what we do when asked to remove a file under the user's TRASH_CAN:
    */
   
   tmp = getenv("TRASH_RECURSE");
   
   if (tmp)
     {
	if (!strcmp(tmp, "UNLINK"))
	  trash_recurse = UNLINK;
	else if (!strcmp(tmp, "NOTHING"))
	  trash_recurse = NOTHING;
	/* If the environmental variable TRASH_RECURSE has any other value, we use the compile-time 
	 * defaulft.*/
     }
   
   
   /* 5- trash_global determines whether files outside of the user's home directory are handled as
    those under the user's home dir (more details available in the configuration section): */
   
   tmp = getenv("TRASH_GLOBAL");
   
   if (tmp)
     {
	if (!strcmp(tmp, "YES"))
	  trash_global = YES;
	else if (!strcmp(tmp, "NO"))
	  trash_global = NO;
     }
   
   /* If trash_global is set, perhaps the user has defined a TRASH_SYSTEM_ROOT (without any
    slashes, we hope :-): */
   
   if (trash_global)
     {
	/* This memory will be free()d below, either after we have built absolute_trash_system_root 
	 * or right away, if it contains a slash: 
	 */
	
	tmp = get_string_env("TRASH_SYSTEM_ROOT");
	
	if (tmp)
	  {
	     if (strchr(tmp, '/'))
	       free(tmp);
	     else
	       relative_trash_system_root = tmp;
	  }
     }
   
   
   /* 6- Is TRASH_TMP_DIRS set? */
   
   /* This memory will be free()d in _fini(): */
   
   tmp = get_string_env("TRASH_TMP_DIRS");
   
   if (tmp)
     tmp_dirs = tmp;
   
   
   /* 7- Is TRASH_UNREM_DIRS set? (We read this now, but unlink() only pays any attention to it
    * if you are root.) */
   
   /* This memory will be free()d in _fini(): */
   
   tmp = get_string_env("TRASH_UNREM_DIRS");
   
   if (tmp)
     unremovable_dirs = tmp;
   
   
   /*  ----------------------------------------------- */
   
   
   /* All the configuration variables are by now set. */
   
   /* Basic sanity test: */
   
   if (!strlen(relative_trash_can) || 
       (trash_global && !strlen(relative_trash_system_root)))
     {
#ifdef DEBUG
	fprintf(stderr, "Intolerable configuration.\nlibtrash turned off.\n");
#endif
	failure = 1;
	return;
     }
   
   
   /*  ----------------------------------------------- */
   
   
   /* Get a pointer to glibc's unlink(): */
   
   unlink_ptr = dlsym(RTLD_NEXT, "unlink");
   
   /* If we didn't get a pointer to the "real" unlink(), we will never be able to use glibc's unlink().
    * This should never happen - we signal failure, set trash_fail to NOTHING (so that we never attempt
    * to run what unlink_ptr points to) and quit:
    */
   
   /* TOTOLOTO  9 10 15 20 33 48 / 42*/
   
   if (!unlink_ptr)
     {
#ifdef DEBUG
	fprintf(stderr, "POSSIBLE MAJOR PROBLEM: unlink_ptr == NULL!\tCalls to the real unlink() - and remove() "
		"- _won't_ work!.\n");
#endif
	trash_fail = NOTHING;
	failure = 1;
	return;
     }
   
   
   
   /* Things left to do:
    * Get information about the user, form two strings containing the absolute
    * pathname of the trash can and the relative_trash_system_root under it (if trash_global
    * is set), see if those dirs exist (try to create it them they don't) and store the user's 
    * login and home directory in memory. 
    */
   
   
   /* I got this from (2) getlogin()'s man page, although they mention geteuid() instead: */
   
   userinfo = getpwuid(getuid());
   
   if (!userinfo)
     {
#ifdef DEBUG
	fprintf(stderr, "Unable to determine necessary data about the user.\nlibtrash turned off.\n");
#endif
	failure = 1;
	return;
     }
   
   
   /* Information which unlink() needs: username, home, absolute_trash_can and
    * absolute_trash_system_root. */
   
   /* This memory will be free()d in _fini() : */
   
   username = malloc(strlen(userinfo->pw_name) + 1);
   home     = malloc(strlen(userinfo->pw_dir) + 1);
   
   absolute_trash_can = malloc(strlen(userinfo->pw_dir) + 1 + strlen(relative_trash_can) + 1);
   
   if (trash_global)
     absolute_trash_system_root = malloc(strlen(userinfo->pw_dir) + 1 + strlen(relative_trash_can)
					 + 1 + strlen(relative_trash_system_root) + 1);
   
   if (!username || !home || !absolute_trash_can || 
       (trash_global && !absolute_trash_system_root))
     {
#ifdef DEBUG
	fprintf(stderr, "Unable to allocate sufficient memory.\nlibtrash turned off.\n");
#endif
	failure = 1;
	return;
     }
   
   strcpy(username, userinfo->pw_name);
   strcpy(home, userinfo->pw_dir);
   
   strcpy(absolute_trash_can, home);
   strcat(absolute_trash_can, "/");
   strcat(absolute_trash_can, relative_trash_can);
   
   if (trash_global)
     {
	strcpy(absolute_trash_system_root, absolute_trash_can);
	strcat(absolute_trash_system_root, "/");
	strcat(absolute_trash_system_root, relative_trash_system_root);
     }
   
   /* This is a good time to free() the memory pointed to by relative_trash_can and relative_trash_system_root, 
    * if they are pointing to dynamically allocated memory. We have already used the information they contain:
    */
   
   if (relative_trash_can != default_relative_trash_can)
     free(relative_trash_can);
   
   if (trash_global && 
       (relative_trash_system_root != default_relative_trash_system_root))
     free(relative_trash_system_root);
   
   
   /* Now we will check the existence and permissions of absolute_trash_can and, if 
    * trash_global is set, absolute_trash_system_root, and create them if they don't 
    * already exist: 
    */
   
   if (!dir_ok(absolute_trash_can, NULL))
     {
#ifdef DEBUG
	fprintf(stderr, "The TRASH_CAN dir (%s) either doesn't exist (and its creation failed) or has insufficient
			  permissions which we were unable to change.\nlibtrash turned off.\n", absolute_trash_can);
#endif
	failure = 1;
	return;
     }
   
   if (trash_global && !dir_ok(absolute_trash_system_root, NULL))
     {
#ifdef DEBUG
	fprintf(stderr, "The TRASH_CAN/SYSTEM_ROOT dir (%s) either doesn't exist (and its creation failed) or has insufficient
			  permissions which we were unable to change.\nlibtrash turned off.\n", absolute_trash_system_root);
#endif
	failure = 1;
	return;
     }
   
   /* We know everything we need to know. _init() is done. */
   
   return;
}


/* -------------------------------  */



/* Just like _init(), this function is automatically invoked, but just before unloading: */

void _fini()
{
   /* The only thing we need to do is free() the dynamically allocated buffers: */
   
   if (absolute_trash_can != NULL)
     free(absolute_trash_can);
   
   if (absolute_trash_system_root != NULL)
     free(absolute_trash_system_root);
   
   if (username != NULL)
     free(username);
   
   if (home != NULL)
     free(home);
   
   if (unremovable_dirs != default_unremovable_dirs)
     free(unremovable_dirs);
   
   if (tmp_dirs != default_tmp_dirs)
     free(tmp_dirs);
   
   return;
}

/* -------------------------------- */




/* What this version of unlink() does: if everything is OK, we just rename() the given file instead of
 unlink()ing it, putting it under absolute_trash_can.  We recreate the entire necessary tree structure under 
 absolute_trash_can. If a file with the same name already exists, the new file will have "[n]" appended to its
 filename, where n is the number of name "collisions" which have occurred for this filename (starting at 1).
 * 
 If the user was trying to unlink() a file outside of her home directory and trash_global is set, we recreate
 the entire necessary tree structure - starting at the root of the filesystem - under 
 absolute_trash_can/trash_system_root. */

int unlink(const char *pathname)
{
   
   struct stat path_stat;
   
   char *absolute_path = NULL;
   
   int retval = 0;
   
   int error = 0;
   
   int jump = 0;
   
   /* If failure is set, we know that something went wrong in _init, and we do whatever trash_fail
    specifies, returning the appropriate error code: */
   
   /*  The following three lines are here to allow helper functions used by unlink() to force it
    * to invoke handle_trash_error() and return its value, should a critical error occur inside them.
    * I think this "centralized" approach to error handling actually makes the code easier to
    * understand than a system relying on having unlink() permanently check the return code of 
    * those functions for "special" values and then invoking handle_trash_error(), specially because
    * most of these functions have their return values interpreted in a boolean way.
    */
   
   jump = setjmp(env); /* Save context / return from longjmp() */
   
   if (jump) /* If returning from longjmp, absolute_path and cwd point to dynamically allocated memory (see
	      * below) and we experienced an error in found_under_dir(). */
     {
#ifdef DEBUG
	fprintf(stderr, "Just arrived from longjmp(), and failure is set.\n");
#endif
	free(absolute_path);
	failure = 1;
     }
   
   
   if (failure)
     {     
#ifdef DEBUG
	fprintf(stderr, "Failure is set in unlink(), invoking handle_trash_error().\ntrash_fail has value %d.\n",
		trash_fail);
#endif
	return handle_trash_error(pathname);
     }
   
   /*  At this point, we can be sure that 
    * 
    * (a) username points to the user's name :-),
    * (b) home points to the user's home dir,
    * (c) absolute_trash_can points to a directory in the user's home directory,
    * (d) to which she has full-permissions,
    * (e) unremovable_dirs lists a list of unremovable directories, which hold files even root shouldn't
    * be able to unlink(), and
    * (f) tmp_dirs lists directories which contain files with which we use the "real" unlink().
    * 
    */
   
   /* First of all: has the user mistakenly asked us to remove a dir or a file which doesn't even exist? 
    * If stat() returns an error code or says it's a dir, we just let the normal unlink() complain about it 
    * and save ourselves the extra trouble:
    */
   
   
   error = stat(pathname, &path_stat);
   
   if (error || S_ISDIR(path_stat.st_mode))
     {
#ifdef DEBUG
	fprintf(stderr, "%s either doesn't exist or is a dir.\nCalling the \"real\" unlink().\n", pathname);
#endif
	return (*unlink_ptr) (pathname);
     }
   
   /* Let us begin by building an absolute path, if we weren't passed one: */
   
   if (pathname[0] != '/')
     {
	
	/* Determine current working directory: */
	
	/* [Will be free()d before after we have built absolute_path.] */
	
	char *cwd = get_current_dir_name();
	
	if (!cwd)
	  {
#ifdef DEBUG
	     fprintf(stderr, "Unable to determine current working directory.\nCall to unlink() will be handled"
		     " according to the value of trash_fail.\n");
#endif
	     return handle_trash_error(pathname);
	  }
	
	
	
	/* Allocate memory necessary for the absolute_path (might add an extra byte for the slash,
	 * if cwd is other than "/"): 
	 */
	
	/* Will be free()d before returning: */
	
	absolute_path = malloc(strlen(cwd) + (strcmp(cwd, "/") ? 1 : 0) + strlen(pathname) + 1);
	
	if (!absolute_path)
	  {
#ifdef DEBUG
	     fprintf(stderr, "Unable to allocate sufficient memory.\nCall to unlink() will be handled"
		     " according to the value of trash_fail.\n");
#endif
	     free(cwd);
	     return handle_trash_error(pathname);
	  }
	
	
	strcpy(absolute_path, cwd);
	
	/* If cwd isn't the root of the file system, append a slash: */
	
	if (strcmp(cwd, "/"))
	  strcat(absolute_path, "/");
	
	strcat(absolute_path, pathname);
	
	/* Free allocated memory at cwd: */
	
	free(cwd);
     }
   else /* if we _were_ passed an absolute path*/
     {
	
	/* Will be free()d before returning: */
	
	absolute_path = malloc(strlen(pathname) + 1);
	
	if (!absolute_path)
	  {
#ifdef DEBUG
	     fprintf(stderr, "Unable to allocate sufficient memory.\nCall to unlink() will be handled"
		     " according to the value of trash_fail.\n");
#endif
	     return handle_trash_error(pathname);
	  }
	
	strcpy(absolute_path, pathname);
     }
   
   
#ifdef DEBUG
   fprintf(stderr, "UNCLEANSED ABSOLUTE_PATH: |%s|\n", absolute_path);
#endif
   
   /* Independently of the way in which the argument was written, absolute_path now holds the absolute
    * path to the file the user wishes to unlink(). We still need to do some more preparation: the path,
    * although absolute, might contain "/../" or "/./", of which we must get rid:
    */
   
   cleanse_path(&absolute_path);
   
#ifdef DEBUG
   fprintf(stderr, "CLEANSED ABSOLUTE_PATH:   |%s|\n", absolute_path);
#endif
   
   /* Point 0.1: handle files already under your TRASH_CAN according to the value of trash_recurse.
    * 
    */
   
   if (found_under_dir(absolute_path, absolute_trash_can))
     {
	free(absolute_path);
#ifdef DEBUG
	fprintf(stderr, "This file is in your TRASH_CAN.\n");
	if (trash_recurse == UNLINK)
	  fprintf(stderr, " allowing real unlink()ing because trash_recurse is set to UNLINK.\n");
	else
	  fprintf(stderr, " disallowing real unlink()ing and returning -1 because trash_recurse is set to NOTHING.\n");
#endif
	if (trash_recurse == UNLINK)
	  return (*unlink_ptr) (pathname);
	else /* if trash_recurse == NOTHING*/
	  return -1;
     }
   
   /* Point 0.2: is this a hidden file or a file under a hidden dir? If so, is trash_ignore_hidden
    * set?
    */
   
   if (trash_ignore_hidden && hidden_file(absolute_path))
     {
	free(absolute_path);
#ifdef DEBUG
	fprintf(stderr, "Ignoring hidden file (or file under hidden dir). Using the real unlink().\n");
#endif
	return (*unlink_ptr) (pathname);
     }
   
   
   
   /* First: are we running as root and is this file in one of the unremovable_dirs? */
   
   /* We check if absolute_path starts with one of the directories listed in unremovable_dirs, and, if it
    * does, we do nothing and return -1:
    */
   
   if (!strcmp(username, "root") && found_under_dir(absolute_path, unremovable_dirs))
     {
#ifdef DEBUG
	fprintf(stderr, "%s is a protected file, returning -1.\n", absolute_path);
#endif
	free(absolute_path);
	return -1;
     }
   
   /* If we are here, this isn't one of the "protected" files - perhaps it is either a temporary file
    * or a file outside of the user's home directory and trash_global isn't set. In any case, invoke
    * the "real" unlink:
    */
   
   if (found_under_dir(absolute_path, tmp_dirs) ||
       (!found_under_dir(absolute_path, home) && !trash_global) )
     {
#ifdef DEBUG
	fprintf(stderr, "%s is either a temporary file or a file outside of your home dir and trash_global isn't"
		" set. Calling the \"real\" unlink().\n", absolute_path);
#endif
	free(absolute_path);
	return (*unlink_ptr) (pathname);
     }
   
   
   /* At this point, we are done with the "exceptions". The remaining possibilities are (a) a file
    * under the user's home dir and (b) a file outside of the user's home dir with trash_global set.
    */
   
   /* (a): */
   
   if (found_under_dir(absolute_path, home))
     {
#ifdef DEBUG
	fprintf(stderr, "%s is under your home dir, normal procedure.\n", absolute_path);
#endif
	retval = graft_file(absolute_trash_can, absolute_path, home);
	free(absolute_path);
	return retval;
     }
   else /* (b) */
     {
#ifdef DEBUG
	fprintf(stderr, "%s isn't under your dir but trash_global is set, normal procedure.\n", absolute_path);
#endif
	retval = graft_file(absolute_trash_system_root, absolute_path, NULL);
	free(absolute_path);
	return retval;
     }
   
   /* Done. */
   
}




/* ---------------------------------------------------------------------- */


/* Helper functions: */

/* ---------------------------------------------------------------------- */



/* This function is called by unlink() in case an error is spotted. It does what the user 
 * told us to do in these situations (through the variable trash_fail), and unlink() just
 * returns the value passed by handle_trash_error():
 */

int handle_trash_error(const char *pathname)
{
   if (trash_fail == UNLINK)
     return (*unlink_ptr) (pathname);
   else
     return -1;
}


/* ------------------------------------------------------------------------ */


/* This function returns 1 if the given path points to a file or directory under one
 * of the directories listed in the semi-colon separated dir_list, 0 otherwise. We use a
 * copy of the buffer pointed to by dir_list because strtok "destroys" the string it is 
 * given as an argument. In case of error (i.e., insufficient memory), we call longjmp()
 * and "return" to the top of the function unlink(), which then invokes handle_trash_error():
 */

static int found_under_dir(const char *abs_path, const char *dir_list)
{
   char *copy_dir_list = NULL, *ptr = NULL;
   
   copy_dir_list = malloc(strlen(dir_list) + 1);
   
   if (!copy_dir_list)
     {
#ifdef DEBUG
	fprintf(stderr, "Unable to allocate sufficient memory.\nlibtrash turned off.\n");
#endif
	longjmp(env, 1); /* I know this is considered "bad practice", but I can't think of a simpler 
			  * solution... If we chose to report errors through the return value (or a 
			  * pointer to an integer), the code of unlink() would need to surround each 
			  * call to found_under_dir() with sanity checks.
			  */
     }
   
   strcpy(copy_dir_list, dir_list);
   
   ptr = strtok(copy_dir_list, ";");
   
   while (ptr)
     {
	if (strstr(abs_path, ptr) == abs_path)
	  {
	     free(copy_dir_list);
	     return 1;
	  }
	ptr = strtok(NULL, ";");
     }
   
   free(copy_dir_list);
   return 0;
   
}


/* ---------------------------------------------------------------------- */


/* This function calls getenv() with the argument it was given and returns either a copy of
 * the string getenv() returned or NULL, in case of failure. This way, if it fails the
 * compile-time default will automatically be used (because we also return NULL if no such
 * variable is set in the environment). This is why we don't call longjmp() in case of 
 * insufficient memory like the other helper functions do: the compile-time defaults are
 * (should be) a valid alternative.
 * 
 * The problem with calling getenv() directly is that the string it returns a pointer to 
 * might be changed by subsequent calls (to getenv()).
 */

char *get_string_env(const char *name)
{
   char *value = NULL, *copy = NULL;
   
   value = getenv(name);
   
   if (!value)
     return NULL;
   
   copy = malloc(strlen(value) + 1);
   
   if (!copy)
     return NULL;
   
   strcpy(copy, value);
   
   return copy;
}

/* --------------------------------------------------------- */

/* This function tests the existence and permissions of the dir's pathname; if it exists and has
 * the right permissions, it returns 1. If it exists but has the wrong permissions and we can't
 * correct them, or if pathname points to a file, it returns 0. If nothing exists at pathname, 
 * it tries to create a dir with full permissions by that name. If that succeeds, it returns 1, 
 * otherwise 0.
 * 
 * Additionally, if it fails due to a name collision (i.e., a file or a dir with that pathname and
 * with wrong - and incorrectable - permissions exists) and the pointer name_collision isn't NULL, 
 * it stores 1 where that pointer points to.
 */


static int dir_ok(const char *pathname, int *name_collision)
{
   int error = 0;
   
   struct stat dir_stat;
   
   /* Run stat() on the given pathname: */
   
   error = stat(pathname, &dir_stat);
   
   
   /* Doesn't a file or dir with such a name exist? */
   
   if (error && (errno == ENOENT))
     {
	/* Try to create it: */
	
	error = mkdir(pathname, S_IRWXU);
	
	/* Did it fail? */
	
	if (error)
	  {
	     if (name_collision)
	       *name_collision = 0;
	     return 0;
	  }
	else
	  return 1;
     }
   else if (error) /* Some other, more obscure, error occurred... just bail out: */
     {
	if (name_collision)
	  *name_collision = 0;
	return 0;
     }
   
   /* By now, we are sure that something with this name exists, but is it a directory? */
   
   if (!S_ISDIR(dir_stat.st_mode))
     {
	/* If the pointer name_collision isn't NULL, this means we should sign that we are returning
	 * 0 because of a name collision: */
	
	if(name_collision)
	  *name_collision = 1;
	return 0;
     }
   
   /* By now, we can be sure that pathname exists and is a directory - but do we have full permissions? */
   
   
   /* [Probably there is a more intelligent way of checking this condition.] */
   
   if ((dir_stat.st_mode & S_IWUSR) && 
       (dir_stat.st_mode & S_IRUSR) && 
       (dir_stat.st_mode & S_IXUSR))
     return 1;
   else
     {
	
	/* Try to change the permissions of that dir: */
	
	error = chmod(pathname, S_IRUSR | S_IWUSR | S_IXUSR);
	
	if (error)
	  {
	     /*
	      * If the pointer name_collision isn't NULL, this means we should sign that we are returning
	      * 0 because of a name collision: */
	     
	     if(name_collision)
	       *name_collision = 1;
	     return 0;
	  }
	else 
	  return 1;
     }
   
   
}

/* -------------------------------------------------------------- */

/* This function reproduces the directory structure described in branch under tree,
 * ignoring the part of branch contained in what_to_cut. It then moves the file at 
 * the end of branch to the end of the newly-created directory hierarchy under tree. 
 * If what_to_cut is NULL, we do not "skip over" any part of the path contained in
 * branch.
 * 
 * E.g.
 * 
 * graft_file("/new_dir", "/a/b/c/file.txt", "/a");
 * 
 * results in rename()ing file "/a/b/c/file.txt" to "/new_dir/b/c/file.txt".
 * 
 * Or:
 * 
 * graft_file("/new_dir", "/a/b/c/file.txt", NULL);
 * 
 * moves file "/a/b/c/file.txt" to "/new_dir/a/b/c/file.txt".
 * 
 * In case of insufficient memory or other serious error, we use longjmp(),
 * although we could also just invoke handle_trash_error() and return its
 * value, because the return value of graft_file() is always returned by
 * unlink(). I chose to use longjmp() for consistency, because this way it
 * is clear that "fatal" allocation issues (and other serious annoyances)
 * inside helper functions are always "solved" at the setjmp() point at the
 * top of unlink(). */


static int graft_file(const char *tree, const char *old_path, const char *what_to_cut)
{
   /* We assume that tree exists and we have full-permissions to it, because it is
    * either absolute_trash_can or absolute_trash_system_root, and their existence
    * and their permissions have been checked by _init().
    * 
    * We assume that the file pointed to by branch resides under what_to_cut (if what_to_cut
    * is NULL, it represents the filesystems root dir, /), because the arguments we were given
    * have been chosen to guarantee that this is so.
    */
   
   char *new_path = NULL;
   
   char *ptr = NULL;
   
   struct stat file_stat;
   
   /* We need to keep the pointer to the original argument intact (so that we can pass it as an
    * argument to rename() or handle_trash_error(), in case any problems come up), so we use
    * branch instead:
    */
   
   const char *branch = old_path;
   
   int error = 0, success = 0;
   
   int name_collision = 0;
   
   
   /* First of all: "cut" what_to_cut from branch. For the above-mentioned reasons, we don't
    * perform any checks, except for testing whether what_to_cut is NULL:
    */
   
   if (what_to_cut)
     branch += strlen(what_to_cut);
   
   /* Now branch points to the relative path to the file which we must reproduce in its entirety
    * under tree. Let's form new_path, which is the concatenation of tree and (the new)
    * branch: 
    */
   
   /* This memory will be free()d before returning: */
   
   new_path = malloc(strlen(tree) + strlen(branch) + 1);
   
   if (!new_path)
     {
#ifdef DEBUG
	fprintf(stderr, "Unable to allocate sufficient memory.\nlibtrash turned off.\n");
#endif
	longjmp(env, 1);
     }
   
   strcpy(new_path, tree);
   strcat(new_path, branch);
   
   /* We proceed in the following way:
    * 
    * (0) point ptr to the character which follows the end of tree in new_path (first char after
    * the slash);
    * 
    * (a) we search for the next slash after ptr, mark it with ptr and overwrite 
    * it with '\0'; if no slash is found, we go to (f);
    * 
    * (b) we feed new_path to dir_ok(): this will test if this directory (whose name is delimited by the
    * '\0' at ptr) already exists (and has the right permissions), and, if it doesn't, dir_ok() will attempt to 
    * correct the situation;
    * 
    * (c) if dir_ok() succeeds, we put the slash back in place (where ptr points to) , move ptr 1 byte 
    * forward and go back to (a);
    * 
    * (d) if dir_ok() failed for a reason other than a name collision, we fail and do a longjmp();
    *
    * (e) if dir_ok() failed due to a name collision, we invoke reformulate_new_path(), which 
    *
    *    i)   points new_path to a viable file path,
    *    ii)  puts the slash back in place and
    *    iii) rewinds ptr so that it points to the beginning of the new, suggested directory name (which
    *         will be created by dir_ok()), instead of pointing at the slash that follows it.
    *
    * We then go back to (a).
    * 
    * (f) we are done, we just need to rename() the file (whose name is stored at old_path) to new_path.
    */
   
   ptr = new_path + strlen(tree) + 1;
   
   while ( (ptr = strchr(ptr, '/')) )
     {
	*ptr = '\0';
	
	success = dir_ok(new_path, &name_collision);
	
	if (!success)
	  {
	     if (!name_collision)
	       {
#ifdef DEBUG
		  fprintf(stderr, "We are quitting graft_file() with a longjmp() because the call to dir_ok()"
			  " failed for a reason other than a name collision.\n"
			  "new_path: %s\n", new_path);
#endif
		  free(new_path);
		  longjmp(env, 1);
	       }
	     
	     else /* if we failed due to a name collision */
	       {
#ifdef DEBUG
		  fprintf(stderr, "dir_ok() failed due to a name collision. Invoking reformulate_new_path.\n"
			  "new_path: %s\n", new_path);
#endif
		  reformulate_new_path(&new_path, &ptr);
		  continue;
	       }
	     
	  }
	
	*ptr = '/';
	ptr++;
     }
   
   /* At this point, we are almost ready; we must only see if there is no collision with the file name
    * itself (i.e., does /a/b/c/file.txt already exist?): */
   
   errno = 0;
   
   error = stat(new_path, &file_stat);
   
   if (!(error && errno == ENOENT))
     reformulate_new_path(&new_path, NULL);
   
   /* The only thing left to do is to rename() the file: */
   
   error = rename(old_path, new_path);
   
#ifdef DEBUG
   fprintf(stderr , "rename() invoked. Arguments passed: old_path: |%s|; new_path: |%s|\n", old_path, new_path);
#endif
   
   /* And free() the memory: */
   
   free(new_path);
   
   /* If the rename() failed, bail out using longjmp() and let unlink() invoke handle_trash_error: */
   
   if (error)
     {
#ifdef DEBUG
	fprintf(stderr, "Doing a longjmp() at the end of graft_file() because rename() returned -1.");
#endif
	longjmp(env, 1);
     }
   else /* Success! :-) */
     return 0;
}


/* ------------------------------------------------------------------------------- */

/* What it does: called if there is a name collision, this function substitutes the "offending"
 * dir or filename with one ressembling the desired one but doesn't already exist. The new file/dir
 * name has the form "original"[n], where n is the lowest natural number which avoids a collision.
 * 
 * If the second argument is NULL, we rewrite the file's name itself. If it isn't, it is assumed to
 * be a pointer to a pointer to the null character which ends the "offending" dir name in the string
 * pointed to by the pointer to which the first argument points. (I think I got this right! :-))
 *
 * In case of an allocation error, we free the memory pointed to by the pointer to which the first
 * argument points and then call longjmp().
 * 
 * [Note: should we limit the number of tries to generate new names?]
 */

static void reformulate_new_path(char **new_path, char **first_null)
{
   
   char *possible_path = NULL;
   
   char *tmp = NULL;
   
   struct stat stat_var;
   
   int error = 0;
   
   int n = 1;
   
   int digits_number = DEF_DIGITS;
   
   /* This memory will be free()d either after having stored a new path in *new_path or
    * inside the calling function (graft_files()), depending on whether first_null is
    * NULL or not.
    */
   
   possible_path = malloc(strlen(*new_path) + 1 + digits_number + 1 + 1);
   
   if (!possible_path)
     {
#ifdef DEBUG
	fprintf(stderr, "Unable to allocate sufficient memory.\n");
#endif
	free(*new_path);
	longjmp(env, 1);
     }
   
   
   errno = 0;
   
   while (!(error && errno == ENOENT))
     {
	sprintf(possible_path, "%s[%d]", *new_path, n++);
	
	error = stat(possible_path, &stat_var);
	
	if (first_null != NULL && /* if we have been asked to invent a new a _dir_ name */
	    !error  &&            /* and if something with this name (possible_path) already exists */
	    S_ISDIR(stat_var.st_mode) && /* and this something is a directory */
	    stat_var.st_mode & S_IWUSR && /* to which you have write-permission */
	    stat_var.st_mode & S_IRUSR && /* and read-permission */
	    stat_var.st_mode & S_IXUSR)  /* and execute-permission */
	  break;                           /* then stop making up new names because this one serves us. */
	
	
	/* Do we need more memory? */
	if ((int)log10(n) == digits_number)
	  {
	     digits_number *= 2;
	     
	     tmp = realloc(possible_path, strlen(*new_path) + 1 + digits_number + 1 + 1);
	     
	     if (!tmp)
	       {
#ifdef DEBUG
		  fprintf(stderr, "Unable to allocate sufficient memory.\n");
#endif
		  free(*new_path);
		  free(possible_path);
		  longjmp(env, 1);
	       }
	     possible_path = tmp;
	  }
     }
   
   /* By now, possible_path holds a "good" replacement for the part of *new_path which goes up to
    * the first '\0', independently of that being the "entire" path or just a part of it. (I mean:
    * if the problem was the name of a directory, we now have an alternative dir name; if the problem
    * was the name of the file itself, we now have a complete new path for that file.)
    * 
    * So, if we were passed NULL as the second argument (which means that we already have all that
    * we need in possible_path), we just:
    */
   
   if (!first_null)
     {
	free(*new_path);
	*new_path = possible_path;
	/* [The memory allocated at possible_path will be free()d in graft_files().] */
	return;
     }
   
   /* If we have in possible_path an alternative dir name to the one held in *new_path up to *first_null,
    * we must now concatenate possible_path with the part of the original path which follows *first_null. 
    */
   else
     {
	char *tmp2 = NULL;
	
	/* Before doing so, we must put the slash back in its place:*/
	
	**first_null = '/';
	
	tmp = malloc(strlen(possible_path) + strlen(*first_null) + 1);
	
	if (!tmp)
	  {
#ifdef DEBUG
	     fprintf(stderr, "Unable to allocate sufficient memory.\n");
#endif
	     free(*new_path);
	     free(possible_path);
	     longjmp(env, 1);
	  }
	
	strcpy(tmp, possible_path);
	strcat(tmp, *first_null);
	
	
	/* Now we have the entire, "corrected" path at tmp. */
	
	/* Update pointer at *first_null: 
	 *
	 * First, we make it point to its original location, the slash
	 * which followed the "offending" dir name: */
	
	*first_null = tmp + strlen(possible_path);
	
	/* [This is a good time to free this memory:] */
	
	free(possible_path);
	
	/* We now wish to point the pointer at *first_null to the character which follows the last 
	 * slash _before_ the one currently pointed to by *first_null.
	 * The slash currently pointed to by *first_null  marks the end
	 * of the suggested new dir name, and the previous one marks its
	 * beginning - precisely what we are interested in:
	 */
	
	/* Temporarily remove the slash which terminates the alternative name: */
	
	**first_null = '\0';
	
	/* With that slash temporarily removed,  find the one which precedes it: */
	
	tmp2 = strrchr(tmp, '/');
	
	/* I think we can be _sure_ that the string to which the pointer to which our first 
	 * argument points is an absolute path, whose top dir is absolute_trash_can, whose
	 * existence has been checked. So, we should also be sure that the dir we are being
	 * asked to invent a name for is under absolute_trash_can. This would mean that the
	 * slash which ends its name _isn't_ the last one in this path, and that there isn't
	 * any reason for tmp2 to be NULL.
	 * 
	 * But, if I am missing something, we won't be taking any chances and just abort:
	 */
	
	
	if (!tmp2)
	  {
#ifdef DEBUG
	     fprintf(stderr, "String manipulation error: unable to find previous slash in path, "
		     "function reformulate_new_path.\n");
#endif
	     free(*new_path);
	     free(tmp);
	     longjmp(env, 1);
	  }
	
	/* Put the last slash back in place: */
	
	**first_null = '/';
	
	
	/* Now just update *first_null: */
	
	*first_null = tmp2 + 1;
	
	/* *first_null now points to the new dir name's first
	 * character, so that, when back in graft_files(), the expression
	 *
	 * ptr = strchr(ptr, '/') (in the test of the while loop)
	 * 
	 * makes ptr point to the last slash, i.e., the one which marks the end of
	 * the new dir name, which will be created dir_ok().
	 */
	
	
	/* Free memory: */
	
	free(*new_path);
	
	/* Update the pointer to which the first argument points, so that it now points to an alternative 
	 * path: 
	 */
	
	/* [This memory will be free()d in graft_file().] */
	
	*new_path = tmp;
	
     }
}

/* --------------------------------------------------------------------------- */

/* This function returns 1 if the specified file is either a hidden file (i.e., its name
 * begins with a dot) or resides under a hidden directory (i.e., under a directory whose
 * name begins with a dot). It returns 0 otherwise.
 * 
 * How we do it: we search for all the occurrences of the substring "/." in the path we were
 * given, and then see if any of them is followed by a character other than a slash or a (second)
 * dot. If this is the case, that "/." marks the beginning of a hidden directory's or file's 
 * name, and we return 1. If no single "/." is found, or all found ones are followed by a slash
 * or a dot, we return 0.
 */ 

static int hidden_file(const char *absolute_path)
{
   
   const char *ptr = absolute_path;
   
   while ((ptr = strstr(absolute_path, "/.")))
     {
	/* By adding 2 to ptr we get a pointer to the char which determines whether this is
	 * a hidden file/dir's name or just a reference to the "/.." or "/./": */
	
	if (*(ptr + 2) == '/' || *(ptr + 2) == '.')
	  continue;
	else
	  return 1;
     }
   
   return 0;
}

/* ----------------------------------------------------------------------------- */

static void cleanse_path(char **original_path)
{
   
   char *s_ptr = NULL, *d_ptr = NULL;
   
   char *tmp = NULL;
   
   /* Set to one if the last copied char was a slash, 0 otherwise: */
   
   int slash = 0;
   
   char *worked_path = malloc(strlen(*original_path) + 1);
   
   if (!worked_path)
     {
#ifdef DEBUG
	fprintf(stderr, "Unable to allocate sufficient memory.\n");
#endif
	longjmp(env, 1);
     }
   
   s_ptr = *original_path;
   d_ptr = worked_path;
   
   while (*s_ptr != '\0')
     {
	if (*s_ptr != '/') /* just a normal character, copy it*/
	  {
	     *d_ptr++ = *s_ptr++;
	     slash = 0;
	     continue;
	  }
	else /* if (*s_ptr == '/') */
	  {
	     if (!slash) /* if the last copied char wasn't a slash, copy this one */
	       {
		  *d_ptr++ = *s_ptr;
		  slash = 1;
	       }
	     if (*(s_ptr + 1) != '.') /* what follows is a "normal" dir name, so just increment s_ptr and continue */
	       {
		  s_ptr++;
		  continue;
	       }
	     else if (*(s_ptr + 2) == '/')  /* "/./", just skip over it to the next slash and continue */
	       {
		  s_ptr += 2;
		  continue;
	       }
	     else if (*(s_ptr + 2) == '.') /* "/../", erase the last dir name up to the previous slash (if any), and continue*/
	       {
		  *(d_ptr - 1) = '\0'; /* used to make strrchr() only find the slash _before_ the one we just wrote*/
		  
		  tmp = strrchr(worked_path, '/');
		  
		  if (!tmp) /* we are at the root of the file system (original_path starts with "/../"), ignore "/.."*/
		    {
		       *(d_ptr - 1) = '/'; /* put the slash back in place, we won't be deleting anything */
		    }
		  else /* we have found the previous slash, which marks the beginning of the name of the dir we should ignore*/
		    {
		       d_ptr = tmp + 1; /* simply rewind d_ptr so that this dir gets overwritten, and top it with a null*/
		       *d_ptr = '\0';
		    }
		  s_ptr += 3;
		  continue;
	       }
	     else /* hidden file / dir name, "/.X"; just increment and continue;*/
	       {
		  s_ptr++;
		  continue;
	       }
	  }
     }
   
   *d_ptr = '\0';
   
   free(*original_path);
   
   /* This memory will be free()d when unlink() returns: */
   *original_path = worked_path;
   
   return;
}
